#!/usr/bin/python3
# -*- coding: utf-8 -*-
import apt
import dbus
import dbus.service
import dbus.mainloop.glib
import os
import time
import datetime
import shutil
import threading
import re
import apt_pkg
import gc
import logging
import subprocess#snap


from apt.debfile import DebPackage
from apt.progress.base import AcquireProgress
from kylindbus.apt_handler import AppActions, AptHandler, WorkitemError
from kylindbus.enums import INTERFACE, INTERFACES, check_apt_lock
from kylindbus.policykit_service import PolicyKitService
from kylindbus.log_manager import logger
from kylindbus.sqlite3_server import Sqlite3Server
from kylindbus.log_message_server import ConnectServerThread
from kylindbus.parse_json import ParseJson

CRONPATH = "/var/spool/anacron/"
SOURCESPATH = "/etc/apt/sources.list.d"
CODEHOME = "/usr/share/kylin-update-manager/"
PACKAGES_MANAGER_TOOLS = 'cn.kylinos.KylinUpdateManager.action'



class WorkItem:
    def __init__(self, pkgname, action, kwargs):
        self.pkgname = pkgname
        self.action = action
        self.kwargs = kwargs


class WorkThread(threading.Thread):
    def __init__(self, dbusDaemon):
        threading.Thread.__init__(self)
        self.dbusDaemon = dbusDaemon
        ("KUM DBUS START ", type(self.dbusDaemon))

    #
    # 函数:线程执行函数
    #
    def run(self):
        while True:
            if len(self.dbusDaemon.worklist) == 0:
                time.sleep(0.2)
                continue
            self.dbusDaemon.mutex.acquire()
            self.item = self.dbusDaemon.worklist.pop(0)
            self.dbusDaemon.mutex.release()
            try:
                func = getattr(self.dbusDaemon.aptHandler, self.item.action)
                if func is None:
                    print("Error action: ", self.item)
                res = func(self.item.pkgname, self.item.kwargs)
                if res is False:
                    print("Action exec failed...")
            except WorkitemError as e:
                # 如果是包列表，此时强制装换为str，后面显示str将出现异常，故而暂不显示包列表
                kwarg = {"apt_appname": str(self.item.pkgname),
                         "apt_percent": str(e.errornum),
                         "error_message": str(e.details),
                         }
                # 安装失败时需要进行二次判断，防止第一次安装失败导致后续的包都安装失败
                result = ""
                results = os.popen("dpkg -l | grep %s" %
                                   self.item.pkgname).readlines()
                for res in results:
                    res = re.sub(" +", " ", res)
                    if str(res).split(" ")[1] == self.item.pkgname:
                        result = res.split(" ")[0]
                if result != "ii":
                    logger.error("安装状态：失败 - 包名：" + str(self.item.pkgname) +
                                 " - 错误码：(" + str(e.errornum) + ") - 详情：" + str(e.details))
                    self.dbusDaemon.kum_apt_signal("apt_error", kwarg)
                    self.dbusDaemon.insert_app_message("apt_error", kwarg)
            except Exception as e:  # 代码异常
                kwarg = {"apt_appname": str(self.item.pkgname),
                         "apt_percent": '24',
                         "error_message": str(e),
                         }
                logger.error("安装状态：失败 - 包名：" + str(self.item.pkgname) +
                             " - 错误码：(24) - 详情：" + str(e), exc_info=True)
                # self.dbusDaemon.kum_apt_signal("apt_error", kwarg)
                # self.dbusDaemon.insert_app_message("apt_error", kwarg)
            finally:
                self.dbusDaemon.error_message = ""
                self.item = None
                self.dbusDaemon.counter -= 1
            time.sleep(0.3)
            self.dbusDaemon.aptHandler.delete_cache()


# acquire progress
class FetchProgress(AcquireProgress):
    def __init__(self, Daemon) -> None:
        super().__init__()
        self.deamon = Daemon

    def pulse(self, owner):
        # partial_present = owner.partial_present
        # total_needed = owner.total_needed
        progress = int((self.current_bytes / self.total_bytes) * 100)
        total_size = self.total_bytes
        # progress = int((partial_present/total_needed)*100)
        # total_size = float(total_needed/(1024**2))
        logging.info("fetch progress:%d,total size:%f" % (progress, total_size))
        self.deamon.DownloadProgressSignal(progress, total_size)
        return

    def stop(self):
        logging.info("fetch progess finished")
        finish_status = False
        if self.current_items == self.total_items:
            finish_status = True
        else:
            finish_status = False
        self.deamon.DownloadFinishSignal(finish_status)
        return

    def fail(self, acquireitemdesc):
        desc = acquireitemdesc.shortdesc
        err_text = acquireitemdesc.owner.error_text
        status = acquireitemdesc.owner.status
        type = ""
        if status == apt_pkg.AcquireItem.STAT_AUTH_ERROR:
            type = "authentication error"
        elif status == apt_pkg.AcquireItem.STAT_ERROR:
            type = "fetch error"
        elif status == apt_pkg.AcquireItem.STAT_TRANSIENT_NETWORK_ERROR:
            type = "network error"
        elif status == apt_pkg.AcquireItem.STAT_IDLE:
            type = "to be fetched"
        elif status == apt_pkg.AcquireItem.STAT_FETCHING:
            type = "fetching"
        elif status == apt_pkg.AcquireItem.STAT_DONE:
            type = "finished"
        else:
            type = "normal"
        self.deamon.DownloadErrorSignal(desc, type, err_text)
        return


class Daemon(PolicyKitService):
    def __init__(self, bus, mainloop):
        self.aptHandler = AptHandler(self)
        self.parse_json = ParseJson(self)
        self.sqlite_server = Sqlite3Server()
        self.worklist = []
        self.cancel_name_list = []
        self.mutex = threading.RLock()
        self.cancelmutex = threading.RLock()
        self.worker_thread = WorkThread(self)
        self.worker_thread.setDaemon(True)
        self.worker_thread.start()
        self.importantwork = False
        self.efi_status = ""
        self.optional_message_list = []
        self.silent_message_list = []
        self.bus_name = dbus.service.BusName(INTERFACE, bus=bus)
        PolicyKitService.__init__(self, self.bus_name, INTERFACES)
        self.counter = 0
        self.mainloop = mainloop
        self.error_message = ""
        self.autoupdate_status = 0
        self.json_list = []


    #
    # 函数：检测计数器是否为0，是否可以退出dbus
    #
    def check_counter(self):
        if self.counter == 0:
            logger.info("自我升级完成，定时自动重启服务")
            self.mainloop.quit()
        else:
            exit_timer = threading.Timer(5, self.check_counter)
            exit_timer.start()

    #
    # 函数：apt安装结果写入数据库
    #
    def insert_app_message(self, type_msg, msg):
        try:
            self.insert_to_sqlite(type_msg, msg, self.optional_message_list, 1)
            self.insert_to_sqlite(type_msg, msg, self.silent_message_list, 0)
        except Exception as e:  # 数据库文件被删除或者新增字段导致需要重新初始化数据库再写入
            self.sqlite_server.init_sqlit()
            self.insert_to_sqlite(type_msg, msg, self.optional_message_list, 1)
            self.insert_to_sqlite(type_msg, msg, self.silent_message_list, 0)
            logger.error("数据库写入异常，包名：%s" % msg["apt_appname"], exc_info=True)

    #
    # 函数：插入app信息到数据库并发送信号
    #
    def insert_to_sqlite(self, type_msg, msg, list, var=1):
        time = datetime.datetime.now()
        timestr = datetime.datetime.strftime(time, "%Y-%m-%d %H:%M:%S")
        n = 0
        for i in list:
            if i["appname"] == msg["apt_appname"]:
                appmsg = list.pop(n)
                icon = None
                if type_msg == "apt_finish":
                    statue = "Success"
                    description = appmsg["description"]
                else:
                    statue = "Fail"
                    if msg.get("error_message") is not None:
                        description = msg["error_message"]
                    else:
                        description = self.error_message

                if msg["apt_percent"] == "100":
                    errorcode = "0"
                else:
                    errorcode = msg["apt_percent"]
                if var == 1:
                    server_errorcode = 10010000 + int(errorcode)
                else:
                    server_errorcode = 10000000 + int(errorcode)
                self.sqlite_server.insert_into_installed(appmsg["appname"], appmsg["source_version"],
                                                         timestr, description, icon,
                                                         statue, i["keyword"], errorcode)
                self.sqlite_server.insert_into_display(
                    'update_time', timestr)  # 写入本地数据库
                if i["keyword"] == "1":  # 可选更新才发送信号用于更新历史界面
                    self.update_sqlite_signal(appmsg["appname"], timestr)
                self.send_message_to_server(
                    appmsg["appname"], appmsg["source_version"], statue, str(server_errorcode))  # 发送到日志服务器
                logger.info("安装状态成功写入数据库，包名：%s" % msg["apt_appname"])
            n += 1

    #
    # 函数：插入deb本地包安装信息到数据库
    #
    def insert_deb_message(self, kum, msg):
        time = datetime.datetime.now()
        timestr = datetime.datetime.strftime(time, "%Y-%m-%d %H:%M:%S")
        if msg["statue"] == "Fail":
            description = msg["error_message"]
        else:
            description = kum["description"]
        try:
            self.sqlite_server.insert_into_installed(kum["appname"], kum["version"],
                                                     timestr, description, kum["icon"],
                                                     msg["statue"], kum["keyword"], msg["errorcode"])
        except:
            self.sqlite_server.init_sqlit()
            self.sqlite_server.insert_into_installed(kum["appname"], kum["version"],
                                                     timestr, description, kum["icon"],
                                                     msg["statue"], kum["keyword"], msg["errorcode"])

    #
    # 函数:新增任务
    # item：任务
    #
    def add_worker_item(self, item):
        print("####add_worker_item:", item)
        self.mutex.acquire()
        self.worklist.append(item)
        self.mutex.release()

    #
    # 函数：获取当前下载包的依赖
    #
    def get_deps_pkg(self, appname):
        url_list = []
        name_list = []
        size_list = []
        fullname_list = []
        all_size = 0
        depends_status = False
        try:
            res = os.popen("LANG= apt-get install --print-uris %s -y" %
                           appname).readlines()
            # print(len(res))
            num = 0
            for i in res:
                if "After this operation," in i:
                    num = num + 1
                    depends_status = True
                    while num < len(res):
                        i = res[num]
                        i_list = i.strip("\n").split(" ")
                        url_list.append(i_list[0].strip("'"))
                        name_list.append(i_list[1].split("_")[0])
                        fullname_list.append(i_list[1])
                        size_list.append(i_list[2])
                        all_size = all_size + int(i_list[2])
                        num = num + 1
                    break
                # 软件包已安装(kylin-update-desktop单独处理)-lj
                elif "0 upgraded," in i and " 0 newly" in i:
                    # print("len:",len(res), "num:", num+1)
                    if len(res) == num + 1:
                        depends_status = True
                        break
                elif "The following packages have unmet dependencies:" in i:
                    logger.error("--%s获取依赖失败--" %
                                 appname + "\n" + "".join(res))
                    depends_status = False
                    break
                num = num + 1
            return url_list, name_list, fullname_list, size_list, all_size, depends_status
        except Exception as e:
            logger.error("依赖解析代码出错--%s--" % appname, exc_info=True)
            return url_list, name_list, fullname_list, size_list, all_size, depends_status

    #
    # 函数：判断efi分区的状态
    #
    def get_efi_status(self):
        with open("/proc/mounts") as f:
            for line in f.readlines():
                line_section = line.split(" ")
                if "/boot/efi" in line_section[1]:
                    status = line_section[3].split(",")
                    for i in status:
                        if i == "ro" or i == "rw":
                            return i
                    break

    #
    # 函数：安装前后设置efi分区的状态为读写
    #
    def set_efi_status(self, before=True):
        try:
            if self.efi_status == "ro" and before == True:
                os.system("mount -o remount,rw /boot/efi")
                os.system("mount -o remount,rw /boot")
            elif self.efi_status == "ro" and before == False:
                os.system("mount -o remount,ro /boot/efi")
                os.system("mount -o remount,ro /boot")
        except Exception as e:
            logger.error("修改/boot/efi挂载状态出错", exc_info=True)

    #
    # 函数：开启线程收集客户端信息发送给php服务器
    #
    def send_message_to_server(self, _appname="", _appversion="", _statue="", _errorcode=""):
        send_thread = ConnectServerThread(self)
        send_thread.get_values(_appname, _appversion, _statue, _errorcode)
        send_thread.start()

    #
    # 函数：共用函数用于设置apt状态并发送信号。
    #
    def send_apt_status(self, name, percent, message, status=True):
        if status:
            kwarg = {"apt_appname": str(name), "apt_percent": str(
                percent), "action": str(message)}
            self.kum_apt_signal("apt_finish", kwarg)
        else:
            kwarg = {"apt_appname": str(name), "apt_percent": str(
                percent), "error_message": str(message)}
            self.kum_apt_signal("apt_error", kwarg)
        self.set_efi_status(False)
        self.counter -= 1

    def send_deb_status(self, name, percent, message, status=True):
        if status:
            kwarg = {"apt_appname": str(name), "apt_percent": str(
                percent), "action": str(message)}
            self.kum_deb_signal("apt_finish", kwarg)
        else:
            kwarg = {"apt_appname": str(name), "apt_percent": str(
                percent), "error_message": str(message)}
            self.kum_deb_signal("apt_error", kwarg)
        self.set_efi_status(False)
        self.counter -= 1
    #
    #函数：设置snap包的安装状态并发送信号。
    #
    def send_snap_status(self, message, status=True):
        if status:
            logger.info("安装成功----------------")
            kwarg = {"action": str(message)}
            self.kum_snap_signal("snap_finish", kwarg)
        else:
            logger.info("安装失败-------------------------")
            kwarg = {"error_message": str(message)}
            self.kum_snap_signal("snap_error", kwarg)
        self.set_efi_status(False)
        self.counter -= 1
    #
    # 函数:初始化cache
    #
    def init_cache(self):
        self.aptHandler.init_cache()

    #
    # 函数:提权
    # polkit()
    #
    def auth_with_policykit(self, lang, text, get_action_id, sender=None):
        # return True
        action = get_action_id
        # lang = os.getenv('LANG')
        if text == "":
            if "zh_CN" in lang:
                text = "要卸载软件"
            else:
                text = "To uninstall the software"
        if not sender:
            raise ValueError('sender == None')
        print("text----",text)
        #        print "auth_with_policykit:", sender
        granted = False
        try:
            obj = dbus.SystemBus().get_object('org.freedesktop.PolicyKit1',
                                              '/org/freedesktop/PolicyKit1/Authority')
            policykit = dbus.Interface(obj, 'org.freedesktop.PolicyKit1.Authority')

            subject = ('system-bus-name', {'name': sender})
            flags = dbus.UInt32(1)  # AllowUserInteraction flag
            if "zh_CN" in lang:
                msg = text + "，您需要验证。"
            else:
                msg = text + "，You need to verify。"
            details = {'polkit.message': msg}
            cancel_id = ''  # No cancellation id
            (granted, notused, details) = policykit.CheckAuthorization(
                subject, action, details, flags, cancel_id)
        except Exception as e:
            print("auth with except......", str(e))
            granted = False

        return granted

    # -------------------------------------dbus methods----------------------------------------

    #
    # 函数:安装和升级
    # install package sa:kum_fetch_signal() and kum_apt_signal()
    #
    @dbus.service.method(INTERFACE, in_signature='s', out_signature='b', sender_keyword='sender')
    def install_and_upgrade(self, pkgName, sender=None):
        print("this is install_and_upgrade", pkgName)
        if ";" in pkgName or " " in pkgName:
            return False
        logger.debug("####install_and_upgrade: ", pkgName)
        self.counter += 1
        item = WorkItem(pkgName, AppActions.INS_AND_UP, None)
        self.add_worker_item(item)

        logger.debug("####install_and_upgrade return")
        return True

    #
    # 函数：创建root目录
    #
    @dbus.service.method(INTERFACE, in_signature='s', out_signature='b', sender_keyword='sender')
    def makedirs(self, path, sender=None):
        self.counter += 1
        path = path.strip()
        path = os.path.abspath(path)
        try:
            if not os.path.exists(path):
                if path == "/var/cache/apt/archives/":
                    os.makedirs(path)
                    self.counter -= 1
                    return True
                else:
                    return False
            else:
                return True
        except:
            self.counter -= 1
            # return False

    #
    # 函数：判断包是否可以安装或可升级
    #
    @dbus.service.method(INTERFACE, in_signature='as', out_signature='as', sender_keyword='sender')
    def check_installed_or_upgrade(self, list, sender=None):
        self.counter += 1
        pkglist = []
        sizelist = []
        self.init_cache()
        for i in list:
            try:
                name = str(i).strip().replace("\n", "")
                app_info = self.parse_json.get_pkg_upgrade_info(name)
                if app_info["json"] == 0:
                    pkg = self.aptHandler.get_pkg_by_name(name)
                    if pkg.is_installed == False or pkg.is_upgradable == True:
                        kum = {"appname": str(pkg.name),
                               "source_version": str(pkg.versions[0]).split("=")[1],
                               "size": str(pkg.candidate.size),
                               "description": str(pkg.candidate.raw_description),
                               "keyword": str(0)}
                        for message in self.silent_message_list:
                            if message["appname"] == kum["appname"]:
                                self.silent_message_list.remove(message)
                                break
                        url_list, name_list, size_list, fullname_list, all_size, depends_status = self.get_deps_pkg(pkg.name)
                        if depends_status:
                            self.silent_message_list.append(kum)
                            pkglist.append(str(pkg.name))
                            sizelist.append(str(pkg.candidate.size))
                elif app_info["json"] == 1:  # json存在但不满足条件直接跳过
                    continue
                else:  # 需要走解析json流程
                    self.json_list.append(str(app_info["package"]))
                    kum = {}
                    kum["appname"] = str(app_info["package"])
                    kum["source_version"] = str(app_info["version"])
                    # kum["size"] = str(pkg.candidate.size)
                    kum["keyword"] = str(1)
                    for app in list:  # 过滤更新列表，删除important.list文件中其它包
                        if app in app_info["install_list"]:
                            app_info["install_list"].remove(app)
                        if app in app_info["upgrade_list"]:
                            app_info["upgrade_list"].remove(app)
                    if not app_info["install_list"] and not app_info["upgrade_list"]:
                        continue
                    app_list = ""
                    for app in app_info["install_list"]:  # 拼接所有包名成字符串
                        app_list = app_list + " " + str(app)
                    for app in app_info["upgrade_list"]:
                        app_list = app_list + " " + str(app)
                    logger.info(str(name) + "包列表:" + app_list)
                    url_list, name_list, size_list, fullname_list, all_size, depends_status = self.get_deps_pkg(
                        app_list)
                    print("all_size---->", all_size)
                    kum["size"] = str(all_size)
                    kum["current_version"] = app_info["version"]
                    if app_info["changelog"] != "":  # 判断changlog是使用changlog字段还是description字段
                        kum["description"] = app_info["changelog"]
                    else:
                        kum["description"] = app_info["description"]
                    kum["install_list"] = app_info["install_list"]
                    kum["upgrade_list"] = app_info["upgrade_list"]
                    pkglist.append(str(app_info["package"]))
                    pkglist.append(str(kum["size"]))
            except:
                logger.error("获取关键更新包信息出错,包名：%s" % str(i), exc_info=True)
                continue
        pkglist.extend(sizelist)
        self.counter -= 1
        return pkglist

    #
    # 函数：获取包的信息
    #
    @dbus.service.method(INTERFACE, in_signature='as', sender_keyword='sender')
    def get_app_message(self, appname_list, sender=None):
        self.init_cache()
        self.counter += 1
        self.parse_json.update_json()
        # self.optional_message_list.clear()
        for i in appname_list:
            kum = {}
            name = str(i).strip().replace("\n", "")  # 防止包名传入\n和空格
            app_info = self.parse_json.get_pkg_upgrade_info(name)
            kum["json"] = app_info["json"]
            print("app info [json] == %s", app_info["json"])
            try:
                try:  # 异常处理，防止正在安装包时重新打开控制面依然显示该包
                    if app_info["json"] != 2:
                        work_pkgname = self.worker_thread.item.pkgname
                        if name == work_pkgname:
                            continue
                except:
                    pass

                if app_info["json"] == 0:  # 不存在json文件
                    pkg = self.aptHandler.get_pkg_by_name(name)
                    kum["appname"] = str(pkg.name)
                    kum["source_version"] = str(pkg.versions[0]).split("=")[1]
                    kum["size"] = str(pkg.candidate.size)
                    kum["keyword"] = str(1)
                    if pkg.is_installed and not pkg.is_upgradable:  # 单包推送已经是最新，直接跳过
                        continue
                    url_list, name_list, size_list, fullname_list, all_size, depends_status = self.get_deps_pkg(
                        pkg.name)
                    if pkg.is_installed:
                        kum["current_version"] = pkg.installed.version
                    else:
                        kum["current_version"] = ""
                    kum["description"] = str(pkg.candidate.raw_description)
                    kum["depends_status"] = depends_status
                    kum["icon"] = ""
                    # 发送信号时，字典kum中的value不能是空list
                    self.important_app_message_signal(kum, url_list, name_list, size_list,
                                                      fullname_list, str(all_size), depends_status)
                    kum["install_list"] = []
                    kum["upgrade_list"] = []
                elif app_info["json"] == 1:  # json存在但不满足条件直接跳过
                    continue
                else:  # 需要走解析json流程
                    self.json_list.append(str(app_info["package"]))
                    kum["appname"] = str(app_info["package"])
                    kum["source_version"] = str(app_info["version"])
                    # kum["size"] = str(pkg.candidate.size)
                    kum["keyword"] = str(1)
                    for app in appname_list:  # 过滤更新列表，删除important.list文件中其它包
                        if app in app_info["install_list"]:
                            app_info["install_list"].remove(app)
                        if app in app_info["upgrade_list"]:
                            app_info["upgrade_list"].remove(app)
                    if not app_info["install_list"] and not app_info["upgrade_list"]:
                        continue
                    app_list = ""
                    for app in app_info["install_list"]:  # 拼接所有包名成字符串
                        app_list = app_list + " " + str(app)
                    for app in app_info["upgrade_list"]:
                        app_list = app_list + " " + str(app)
                    logger.info(str(name) + "包列表:" + app_list)
                    url_list, name_list, size_list, fullname_list, all_size, depends_status = self.get_deps_pkg(
                        app_list)
                    print("all_size---->", all_size)
                    kum["size"] = str(all_size)
                    # if pkg.is_installed:
                    #     kum["current_version"] = pkg.installed.version
                    # else:
                    kum["current_version"] = app_info["version"]
                    if app_info["changelog"] != "":  # 判断changlog是使用changlog字段还是description字段
                        kum["description"] = app_info["changelog"]
                    else:
                        kum["description"] = app_info["description"]
                    kum["depends_status"] = depends_status
                    kum["icon"] = app_info["icon"]
                    # 发送信号时，字典kum中的value不能是空list
                    self.important_app_message_signal(kum, url_list, name_list, size_list,
                                                      fullname_list, str(all_size), depends_status)
                    kum["install_list"] = app_info["install_list"]
                    kum["upgrade_list"] = app_info["upgrade_list"]
                    print("--->", kum["install_list"])
                for message in self.optional_message_list:
                    if message["appname"] == kum["appname"]:
                        self.optional_message_list.remove(message)
                        break
                self.optional_message_list.append(kum)
            except Exception as e:
                logger.error("获取重要更新包信息出错,包名：%s 原因: %s" % (str(name), str(e)))
                continue
        self.get_message_finished_signal(str(len(self.optional_message_list)))
        self.counter -= 1
        del pkg
        del self.parse_json.cache
        gc.collect()
        self.aptHandler.delete_cache()

    #
    # dbus接口：向数据库中插入数据
    #
    @dbus.service.method(INTERFACE, in_signature='ss', sender_keyword='sender')
    def insert_install_state(self, item, value, sender=None):
        self.counter += 1
        try:
            self.sqlite_server.insert_into_display(item, value)
            self.counter -= 1
            logger.info("数据库状态表插入成功，选项：%s，值：%s" % (item, value))
        except Exception as e:
            self.counter -= 1
            logger.error("数据库状态表插入失败，选项：%s，值：%s" % (item, value))

    #
    # dbus接口：设置自动更新状态
    #
    @dbus.service.method(INTERFACE, in_signature='b', sender_keyword='sender')
    def set_autoupdate_state(self, status, sender=None):
        self.counter += 1
        if status:
            self.autoupdate_status = 1
            self.counter -= 1
            logger.info("自动更新已开始")
        else:
            self.autoupdate_status = 0
            self.counter -= 1
            logger.info("自动更新已结束")

    #
    # dbus接口：判断包是否可以安装或可升级
    #
    @dbus.service.method(INTERFACE, out_signature='b', sender_keyword='sender')
    def clear_install_list(self, sender=None):
        self.counter += 1
        if self.autoupdate_status:
            logger.info("自动更新正在进行中，暂不允许清理下载列表")
            return False
        try:
            self.mutex.acquire()
            self.worklist.clear()
            self.mutex.release()
            self.counter -= 1
            logger.info("成功清理下载列表")
            return True
        except Exception as e:
            logger.error("清理下载列表失败", exc_info=True)
            self.counter -= 1
            return False

    #
    # dbus接口：wget下载软件包后拷贝到/var/cache/apt/archives目录下
    #
    @dbus.service.method(INTERFACE, in_signature='ass', out_signature='')
    def copy_file_to_install(self, src_list, appname):
        self.counter += 1
        for i in src_list:
            if os.path.exists(os.path.join("/var/cache/apt/archives", os.path.basename(i))):
                pass
            else:
                shutil.move(i, os.path.join(
                    "/var/cache/apt/archives", os.path.basename(i)))
        self.copy_finish(appname)
        self.counter -= 1

    #
    # dbus接口：设置是否正在进行关键更新的状态
    #
    @dbus.service.method(INTERFACE, in_signature='b', out_signature='b', sender_keyword='sender')
    def set_important_status(self, status, sender=None):
        self.counter += 1
        try:
            self.importantwork = status
        except:
            self.counter -= 1
            return False
        self.counter -= 1
        return True

    #
    # dbus接口：获取是否正在关键更新的状态
    #
    @dbus.service.method(INTERFACE, out_signature='b', sender_keyword='sender')
    def get_important_status(self, sender=None):
        return self.importantwork

    #
    # dbus接口：安装本地deb包
    #
    @dbus.service.method(INTERFACE, in_signature='s', out_signature='b', sender_keyword='sender')
    def install_debfile(self, path, sender=None):
        # print("####install deb file: ", path)
        logger.info("开始安装----- %s -----", path)
        self.counter += 1
        self.efi_status = self.get_efi_status()
        self.set_efi_status(True)
        self.aptHandler.init_cache()
        # print("send auth signal")
        # 阶段一：检查路径是否存在
        path = str(path).strip().replace("\n", "")  # 防止路径传入\n和空格
        try:
            debfile = DebPackage(path, self.aptHandler.cache)
        except Exception as e:
            self.send_deb_status(path, 90, str(e), False)
            logger.error("安装状态：失败 - 路径：" + path + " - 错误码：(90) - 详情：该路径不存在")
            del path
            gc.collect()
            return False
        # 阶段二：调用apt模块解析依赖并获取依赖
        kum = {"appname": debfile._sections["Package"],
               "version": debfile._sections["Version"],
               "description": debfile._sections["Description"],
               "icon": "",
               "keyword": str(2)}
        debfile.check()  # do debfile.check for the next to do debfile.missing_deps
        # print("len:%s" % len(debfile.missing_deps))
        # 安装源中存在的依赖包
        if len(debfile.missing_deps) != 0:
            logger.info("开始解析依赖，需要的依赖为", debfile.missing_deps)
            self.aptHandler.init_cache()
            for deb in debfile.missing_deps:
                try:
                    pkg = self.aptHandler.get_pkg_by_name(deb)
                except Exception as e:
                    self.send_deb_status(deb, 99, e, False)
                    msg = {"statue": "Fail", "errorcode": str(99), "error_message": "安装本地包无法获取依赖%s" % deb}  # 记录结果写入数据库
                    self.insert_deb_message(kum, msg)
                    logger.error("安装状态：失败 - 路径：" + path + " - 错误码：(99) - 详情：安装本地包无法获取依赖%s" % deb)
                    del path, debfile
                    gc.collect()
                    return False
                pkg.mark_install()
                try:
                    self.aptHandler.cache.commit()
                except apt.cache.LockFailedException as e:
                    msg = {"statue": "Fail", "errorcode": str(98), "error_message": "dpkg被锁导致安装依赖包失败"}  # 记录结果写入数据库
                    self.insert_deb_message(kum, msg)
                    self.send_deb_status(deb, 98, e, False)
                    logger.error("安装状态：失败 - 路径：" + path + " - 错误码：(98) - 详情：dpkg被锁导致安装依赖包失败")
                    del path, debfile
                    gc.collect()
                    return False
                except SystemError as e:
                    msg = {"statue": "Fail", "errorcode": str(90), "error_message": str(e)}  # 记录结果写入数据库
                    self.insert_deb_message(kum, msg)
                    self.send_deb_status(deb, 93, e, False)
                    logger.error("安装状态：失败 - 路径：" + path + " - 错误码：(90) - 详情：%s" % str(e))
                    del path, debfile
                    gc.collect()
                    return False
        else:
            logger.info("不需要依赖")

        # 阶段三：安装软件包
        # res = debfile.install(AptProcess(self, pkgName, AppActions.INSTALL_ONE,1,pkgName))
        try:
            logger.info("开始安装软件包：", debfile._sections["Package"])
            res = debfile.install()
            if res:  # 安装失败
                logger.info("安装软件包失败--->", debfile._sections["Package"])
                try:
                    self.aptHandler.init_cache()
                    pkg = self.aptHandler.get_pkg_by_name(kum["appname"])
                    cmd = "dpkg -l {} | cut -d ' ' -f 1  | tail -1".format(debfile._sections["Package"])
                    r = os.popen(cmd)
                    text = r.read()
                    r.close()
                    text = text.strip().replace("\n", "")
                    if pkg.is_installed and text != "ii":
                        pkg.mark_delete(True, True)  # 安装时失败后需要直接卸载防止apt出现依赖问题
                        self.aptHandler.cache.commit(None, None)
                except:
                    pass
                msg = {"statue": "Fail", "errorcode": str(97), "error_message": "deb包存在问题导致安装失败"}  # 记录结果写入数据库
                self.insert_deb_message(kum, msg)
                self.send_deb_status(kum["appname"], 97, "deb包存在问题导致安装失败", False)
                logger.error("安装状态：失败 - 路径：" + path + " - 错误码：(97) - 详情：deb包存在问题导致安装失败")
                del path, debfile
                gc.collect()
                return False
            else:  # 安装成功
                msg = {"statue": "Success", "errorcode": str(0), "error_message": ""}  # 记录结果写入数据库
                self.insert_deb_message(kum, msg)
                self.send_deb_status(kum["appname"], 100, AppActions.INSTALL_DEBFILE)
                logger.info("安装状态：成功 - 路径：" + path)
        except apt_pkg.Error as error:
            msg = {"statue": "Fail", "errorcode": str(96), "error_message": str(error)}  # 记录结果写入数据库
            self.insert_deb_message(kum, msg)
            check_apt_lock()  # dpkg、apt解锁
            self.send_deb_status(kum["appname"], 96, str(error), False)
            logger.info("安装状态：失败 - 路径：" + path + " - 错误码：(96) - 详情：%s" % str(error))
            del path, debfile
            gc.collect()
            return False
        except SystemError as e:
            msg = {"statue": "Fail", "errorcode": str(80), "error_message": str(e)}  # 记录结果写入数据库
            self.insert_deb_message(kum, msg)
            self.send_deb_status(kum["appname"], 80, str(e), False)
            logger.info("安装状态：失败 - 路径：" + path + " - 错误码：(80) - 详情：%s" % str(e))
            del path, debfile
            gc.collect()
            return False
        except Exception as e:
            msg = {"statue": "Fail", "errorcode": str(79), "error_message": str(e)}  # 记录结果写入数据库
            self.insert_deb_message(kum, msg)
            self.send_deb_status(kum["appname"], 79, str(e), False)
            logger.info("安装状态：失败 - 路径：" + path + " - 错误码：(79) - 详情：%s" % str(e))
            del path, debfile
            gc.collect()
            return False
        except BaseException as e:
            msg = {"statue": "Fail", "errorcode": str(78), "error_message": str(e)}  # 记录结果写入数据库
            self.insert_deb_message(kum, msg)
            self.send_deb_status(kum["appname"], 78, str(e), False)
            logger.info("安装状态：失败 - 路径：" + path + " - 错误码：(78) - 详情：%s" % str(e))
            del path, debfile
            gc.collect()
            return False
        except:
            logger.info("安装的execpt函数")
        logger.info("安装结束----- %s -----" % path)
        del path, debfile
        gc.collect()
        return True
    #
    # dbus接口：安装本地snap包
    #
    @dbus.service.method(INTERFACE, in_signature='s', out_signature='b', sender_keyword='sender')
    def install_snap(self, path, sender=None):
        logger.info("开始安装snap --------------------%s-----------------" % path)
        self.counter += 1
        self.efi_status = self.get_efi_status()
        self.set_efi_status(True)
        # print("send auth signal")
        if not os.path.exists(path):
            logger.info("安装失败------路径不存在--------------%s-----------------" % path)
            msg = {"statue": "Fail", "errorcode": str(0), "error_message": "路径不存在"}
            self.send_snap_status(msg,False);
            return False
        try:
            status= subprocess.getstatusoutput("snap install " +path+" --dangerous")
        except:
             msg = {"statue": "Fail", "errorcode": str(0), "error_message": " 命令执行报错有问题"}
             self.send_snap_status(msg,False);
        if status[0]==0:
            logger.info("安装成功----------------%s---------" % path)
            msg = {"statue": "Success", "errorcode": str(0), "error_message": ""}
            self.send_snap_status(msg,True)
            return True
        else:
            logger.info("安装失败--------------------%s-----------------" % path)
            if "supported architectures" in status[1]:
                logger.info("架构不符--------------------%s-----------------" % path)
                msg = {"statue": "Fail", "errorcode": str(0), "error_message": "架构不符"}
                self.send_snap_status(msg,False);
                return False
            else:
                logger.info("其他错误--------------------%s-----------------" % path)
                msg = {"statue": "Fail", "errorcode": str(0), "error_message": "其他错误"}
                self.send_snap_status(msg,False);
                return False

    #
    # dbus接口：卸载软件包
    #
    @dbus.service.method(INTERFACE, in_signature='ss', out_signature='b', sender_keyword='sender')
    def remove(self, pkgName, langstr, sender=None):
        # print("####remove: ", pkgName)
        # 阶段一：初始化模块，获取软件信息

        logger.info("开始卸载----- %s -----%s", pkgName, langstr)
        granted = self.auth_with_policykit(langstr, "", "cn.kylinos.KylinUpdateManager.action", sender)
        if not granted:
            print("提权失败")
            self.send_deb_status(pkgName, 90, str("提权失败"), False)
            return False
        self.kum_deb_remove()
        self.counter += 1
        self.efi_status = self.get_efi_status()
        self.set_efi_status(True)
        self.aptHandler.init_cache()
        pkgName = str(pkgName).strip().replace("\n", "")
        try:
            pkg = self.aptHandler.get_pkg_by_name(pkgName)
        except Exception as e:
            self.send_deb_status(pkgName, 95, str(e), False)
            logger.error("卸载状态：失败 - 包名：" + pkgName + " - 错误码：(95) - 详情：%s" % str(e))
            return False
        if pkg.is_installed is False:
            error_message = "%s未安装" % pkg
            self.send_deb_status(pkgName, 94, error_message, False)
            logger.error("卸载状态：失败 - 包名：" + pkgName + " - 错误码：(94) - 详情：软件包未安装")
            return True
        # 阶段二：开始卸载软件包
        pkg.mark_delete(True, True)
        try:
            self.aptHandler.cache.commit(None, None)
        except apt.cache.LockFailedException as e:
            self.send_deb_status(pkgName, 93, str(e), False)
            logger.error("卸载状态：失败 - 包名：" + pkgName + " - 错误码：(93) - 详情：%s" % str(e))
            # raise WorkitemError(3, "package manager is running.")
        except Exception as e:
            os.system("dpkg --configure -a")
            self.send_deb_status(pkgName, 92, str(e), False)
            logger.error("卸载状态：失败 - 包名：" + pkgName + " - 错误码：(92) - 详情：%s" % str(e))
        else:
            self.send_deb_status(pkgName, 100, AppActions.REMOVE_DEBFILE)
        logger.info("卸载结束----- %s -----", pkgName)
        return True

    #
    # dbus接口：开启定时器去检测当前dbus正在被使用，没有被使用后立刻退出
    #
    @dbus.service.method(INTERFACE)
    def regular_check(self):
        exit_timer = threading.Timer(5, self.check_counter)
        exit_timer.start()

    #
    # dbus接口：控制面板退出时清除cache
    #
    @dbus.service.method(INTERFACE)
    def quitDeleteAptCache(self):
        pass
        # self.aptHandler.delete_cache()

    # download certain package and its dependencies
    @dbus.service.method(INTERFACE, in_signature='si', out_signature='b', sender_keyword='sender')
    def DownloadPackages(self, pkgname, try_times, sender=None):
        logging.debug("download packages")
        if try_times > 0:
            pass
        else:
            try_times = 1
        times = try_times
        while times > 0:
            logging.debug("download try time:%d" % times)
            apt_pkg.init()
            cache = apt_pkg.Cache()
            depcache = apt_pkg.DepCache(cache)
            packagerecords = apt_pkg.PackageRecords(cache)
            packagemanager = apt_pkg.PackageManager(depcache)
            sourcelist = apt_pkg.SourceList()
            sourcelist.read_main_list()
            try:
                pkg = cache[pkgname]
                depcache.mark_install(pkg)
            except Exception:
                return False
            fetcher = apt_pkg.Acquire(FetchProgress(self))
            packagemanager.get_archives(fetcher, sourcelist, packagerecords)
            ret = fetcher.run()
            remote_item_count = 0
            local_item_count = 0
            for item in fetcher.items:
                if item.local:
                    local_item_count += 1
                else:
                    remote_item_count += 1
            logging.debug("%d,%d" % (remote_item_count, local_item_count))
            if ret == apt_pkg.Acquire.RESULT_CONTINUE and remote_item_count == 0:
                return True
            times -= 1
        return False

    #
    # dbus接口：改变apt下载速度
    #
    @dbus.service.method(INTERFACE, in_signature='sb')
    def set_downloadspeed_max(self, speed, set):
        if set:
            with open("/etc/apt/apt.conf.d/80apt-download", "w+") as f:
                text = f.write("Acquire::http::Dl-Limit" + " \"" + "%s" % str(speed) + "\";\n")
                text = f.write("Acquire::https::Dl-Limit" + " \"" + "%s" % str(speed) + "\";\n")
                return True
        # else:
        #     if os.path.exists("/etc/apt/apt.conf.d/80apt-download"):
        #         os.remove("/etc/apt/apt.conf.d/80apt-download")
        #         return False
        #     else:
        #         pass
        else:  
            with open("/etc/apt/apt.conf.d/80apt-download", "w+") as f:
                text = f.write("Acquire::http::Dl-Limit" + " \"0\";\n")
                text = f.write("Acquire::https::Dl-Limit" + " \"0\";\n")
                return False

    #
    # dbus接口：获取apt下载速度
    #
    @dbus.service.method(INTERFACE, out_signature='bs')
    def get_downloadspeed_limit_value(self):
        try:
            f = open("/etc/apt/apt.conf.d/80apt-download", "r")
            value = f.readline().split()
            return True, int(value[1][1:-2])
        except:
            return False, 0

    # -------------------------------------dbus signals----------------------------------------
    #
    # dbus信号：用于发送apt下载时进度信息
    #
    @dbus.service.signal(INTERFACE, signature='sa{sv}')
    def kum_fetch_signal(self, type_msg, msg):
        pass

    #
    # dbus信号：用于发送apt安装时进度信息
    #
    @dbus.service.signal(INTERFACE, signature='sa{sv}')
    def kum_apt_signal(self, type_msg, msg):
        pass

    #
    # dbus 信号：用于发送本地安装卸载进度信息（不与更新混用）
    #
    @dbus.service.signal(INTERFACE, signature='sa{sv}')
    def kum_deb_signal(self, type_msg, msg):
        pass

    #
    # dbus 信号：用于发送本地安装snap包的结果
    #
    @dbus.service.signal(INTERFACE, signature='sa{sv}')
    def kum_snap_signal(self, type_msg, msg):
        pass

    #
    # dbus 信号：用于发送卸载中信号
    #
    @dbus.service.signal(INTERFACE, signature='')
    def kum_deb_remove(self):
        pass

    #
    # dbus信号：用于拷贝包到/var/cache/apt/archives的完成信号
    #
    @dbus.service.signal(INTERFACE, signature="s")
    def copy_finish(self, appname):
        pass

    #
    # dbus信号：发送重要更新包列表信息
    #
    @dbus.service.signal(INTERFACE, signature='a{sv}asasasassb')
    def important_app_message_signal(self, msg, url_list, name_list, size_list, fullname_list, all_size,
                                     depends_status):
        pass

    #
    # dbus信号：发送关键更新包列表信息
    #
    @dbus.service.signal(INTERFACE, signature='a{sv}')
    def crucial_app_message_signal(self, msg):
        pass

    #
    # dbus信号：通知前端获取重要更新信息获取完成的dbus信号
    #
    @dbus.service.signal(INTERFACE, signature='s')
    def get_message_finished_signal(self, number):
        pass

    #
    # dbus信号：通知前端数据库写入完成后的dbus信号，用于实时刷新历史更新界面
    #
    @dbus.service.signal(INTERFACE, signature='ss')
    def update_sqlite_signal(self, appname, time):
        pass

    # dbus signal:send the progress info of certain packages download
    @dbus.service.signal(INTERFACE, signature='ii')
    def DownloadProgressSignal(self, progress, total_size):
        pass

    # dbus signal:send the finish signal after each certain packages download
    @dbus.service.signal(INTERFACE, signature='b')
    def DownloadFinishSignal(self, status):
        pass

    # dbus signal:send the error text when there're errors during download
    @dbus.service.signal(INTERFACE, signature='sss')
    def DownloadErrorSignal(self, desc, type, err_text):
        pass
